@@ -1,6 +1,3 @@
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE ForeignFunctionInterface #-}
-
 -- Copyright (C) 2010 John Millikin <jmillikin@gmail.com>
 --
 -- This program is free software: you can redistribute it and/or modify
@@ -22,13 +19,13 @@ module Network.Protocol.SASL.GNU
 	  headerVersion
 	, libraryVersion
 	, checkVersion
-	
+
 	-- * SASL Contexts
 	, SASL
 	, runSASL
 	, setCallback
 	, runCallback
-	
+
 	-- * Mechanisms
 	, Mechanism (..)
 	, clientMechanisms
@@ -36,33 +33,33 @@ module Network.Protocol.SASL.GNU
 	, clientSuggestMechanism
 	, serverMechanisms
 	, serverSupports
-	
+
 	-- * SASL Sessions
 	, Session
 	, runClient
 	, runServer
 	, mechanismName
-	
+
 	-- ** Session Properties
 	, Property (..)
 	, setProperty
 	, getProperty
 	, getPropertyFast
-	
+
 	-- ** Session IO
 	, Progress (..)
 	, step
 	, step64
 	, encode
 	, decode
-	
+
 	-- ** Error handling
 	, Error (..)
 	, catch
 	, handle
 	, try
 	, throw
-	
+
 	-- * Bundled codecs
 	, toBase64
 	, fromBase64
@@ -77,9 +74,11 @@ module Network.Protocol.SASL.GNU
 -- Imports {{{
 
 import           Prelude hiding (catch)
-import           Control.Applicative (Applicative, pure, (<*>))
+import           Data.Maybe (fromMaybe)
+import           Control.Applicative (Applicative, pure, (<*>), (<$>))
 import qualified Control.Exception as E
-import           Control.Monad (ap, when, unless)
+import           Control.Monad (ap, when, unless, (<=<))
+import           Control.Monad.Loops (unfoldrM)
 import           Control.Monad.IO.Class (MonadIO, liftIO)
 import qualified Control.Monad.Trans.Reader as R
 import qualified Data.ByteString as B
@@ -87,7 +86,6 @@ import qualified Data.ByteString.Unsafe as B
 import qualified Data.ByteString.Char8 as Char8
 import           Data.Char (isDigit)
 import           Data.String (IsString, fromString)
-import           Data.Typeable (Typeable)
 import qualified Foreign as F
 import qualified Foreign.C as F
 import           System.IO.Unsafe (unsafePerformIO)
@@ -121,10 +119,9 @@ libraryVersion = io where
 	io = do
 		cstr <- gsasl_check_version F.nullPtr
 		maybeStr <- F.maybePeek F.peekCString cstr
-		return $ case maybeStr >>= parseVersion of
-			Just version -> version
-			Nothing -> error $ "Invalid version string: " ++ show maybeStr
-	
+		return $ fromMaybe (error $ "Invalid version string: " ++ show maybeStr)
+			(maybeStr >>= parseVersion)
+
 	eof = do
 		s <- P.look
 		unless (null s) P.pfail
@@ -144,8 +141,8 @@ instance Functor SASL where
 	fmap f = SASL . fmap f . unSASL
 
 instance Applicative SASL where
-        pure = SASL . pure
-        (<*>) = ap
+	pure = SASL . pure
+	(<*>) = ap
 
 instance Monad SASL where
 	return = SASL . return
@@ -210,9 +207,9 @@ clientSuggestMechanism :: [Mechanism] -> SASL (Maybe Mechanism)
 clientSuggestMechanism mechs = do
 	let bytes = B.intercalate (Char8.pack " ") [x | Mechanism x <- mechs]
 	ctx <- getContext
-	liftIO $ B.useAsCString bytes $ \pMechlist ->
-		gsasl_client_suggest_mechanism ctx pMechlist >>=
-		F.maybePeek (fmap Mechanism . B.packCString)
+	liftIO $ B.useAsCString bytes $
+		F.maybePeek (fmap Mechanism . B.packCString) <=<
+		gsasl_client_suggest_mechanism ctx
 
 -- | A list of 'Mechanism's supported by the @libgsasl@ server.
 serverMechanisms :: SASL [Mechanism]
@@ -255,8 +252,8 @@ instance Functor Session where
 	fmap f = Session . fmap f . unSession
 
 instance Applicative Session where
-        pure = Session . pure
-        (<*>) = ap
+	pure = Session . pure
+	(<*>) = ap
 
 instance Monad Session where
 	return = Session . return
@@ -273,15 +270,15 @@ runSession start (Mechanism mech) session = bracketSASL newSession freeSession i
 		B.useAsCString mech $ \pMech ->
 		F.alloca $ \pSessionCtx -> E.handle noSession $ do
 		start ctx pMech pSessionCtx >>= checkRC
-		fmap (Right . SessionCtx) $ F.peek pSessionCtx
+		Right . SessionCtx <$> F.peek pSessionCtx
 	noSession (SASLException err) = return $ Left err
-	
+
 	freeSession (Left _) = return ()
 	freeSession (Right (SessionCtx ptr)) = gsasl_finish ptr
-	
+
 	io (Left err) = return $ Left err
 	io (Right sctx) = E.catch
-		(fmap Right $ R.runReaderT (unSession session) sctx)
+		(Right <$> R.runReaderT (unSession session) sctx)
 		(\(SASLException err) -> return $ Left err)
 
 -- | Run a session using the @libgsasl@ client.
@@ -335,7 +332,7 @@ data Error
 	| NoPIN
 	| NoService
 	| NoHostname
-	
+
 	| GSSAPI_ReleaseBufferError
 	| GSSAPI_ImportNameError
 	| GSSAPI_InitSecContextError
@@ -350,10 +347,10 @@ data Error
 	| GSSAPI_InquireMechForSASLNameError
 	| GSSAPI_TestOIDSetMemberError
 	| GSSAPI_ReleaseOIDSetError
-	
+
 	| KerberosV5_InitError
 	| KerberosV5_InternalError
-	
+
 	| SecurID_ServerNeedAdditionalPasscode
 	| SecurID_ServerNeedNewPIN
 
@@ -367,8 +364,7 @@ instance Show Error where
 strError :: Error -> String
 strError err = unsafePerformIO $ gsasl_strerror (cFromError err) >>= F.peekCString
 
-data SASLException = SASLException Error
-	deriving (Show, Typeable)
+newtype SASLException = SASLException Error deriving (Show)
 
 instance E.Exception SASLException
 
@@ -427,7 +423,7 @@ cToError x = case x of
 	35 -> NoClientCode
 	36 -> NoServerCode
 	51 -> NoCallback
-	52 -> NoAnonymousToken 
+	52 -> NoAnonymousToken
 	53 -> NoAuthID
 	54 -> NoAuthzID
 	55 -> NoPassword
@@ -493,7 +489,7 @@ data Property
 	| PropertyScramIter
 	| PropertyScramSalt
 	| PropertyScramSaltedPassword
-	
+
 	| ValidateSimple
 	| ValidateExternal
 	| ValidateAnonymous
@@ -569,7 +565,7 @@ getProperty prop = do
 	liftIO $ do
 		cstr <- gsasl_property_get sctx (cFromProperty prop)
 		if cstr /= F.nullPtr
-			then fmap Just $ B.packCString cstr
+			then Just <$> B.packCString cstr
 			else do
 				liftIO $ checkCallbackException sctx
 				return Nothing
@@ -612,28 +608,28 @@ freeCallbackHook ptr = unless (ptr == F.nullPtr) $ do
 callbackImpl :: (Property -> Session Progress) -> CallbackFn
 callbackImpl cb _ sctx cProp = let
 	globalIO = error "globalIO is not implemented"
-	
+
 	sessionIO = do
 		let session = cb $ cToProperty cProp
-		fmap cFromProgress $ R.runReaderT (unSession session) (SessionCtx sctx)
-	
+		cFromProgress <$> R.runReaderT (unSession session) (SessionCtx sctx)
+
 	onError :: SASLException -> IO F.CInt
 	onError (SASLException err) = return $ cFromError err
-	
+
 	onException :: E.SomeException -> IO F.CInt
 	onException exc = do
 		-- A bit ugly; session hooks aren't used anywhere else in
 		-- the binding, so the exception is stashed here.
 		stablePtr <- F.newStablePtr exc
 		gsasl_session_hook_set sctx $ F.castStablePtrToPtr stablePtr
-		
+
 		-- standard libgsasl return codes are all >= 0, so using -1
 		-- provides an easy way to determine later whether the
 		-- exception came from Haskell code.
 		return (-1)
-	
+
 	catchErrors io = E.catches io [E.Handler onError, E.Handler onException]
-	
+
 	in catchErrors $ if sctx == F.nullPtr then globalIO else sessionIO
 
 foreign import ccall "wrapper"
@@ -699,10 +695,10 @@ step input = bracketSession get free peek where
 		rc <- gsasl_step sctx pInput (fromIntegral inputLen) pOutput pOutputLen
 		when (rc /= 0) $ checkCallbackException sctx
 		progress <- checkStepRC rc
-		cstr <- F.peek pOutput
 		cstrLen <- F.peek pOutputLen
+		cstr <- F.peek pOutput
 		return (cstr, cstrLen, progress)
-	
+
 	free (cstr, _, _) = gsasl_free cstr
 	peek (cstr, cstrLen, progress) = do
 		output <- B.packCStringLen (cstr, fromIntegral cstrLen)
@@ -720,7 +716,7 @@ step64 input = bracketSession get free peek where
 		progress <- checkStepRC rc
 		cstr <- F.peek pOutput
 		return (cstr, progress)
-	
+
 	free (cstr, _) = gsasl_free cstr
 	peek (cstr, progress) = do
 		output <- B.packCString cstr
@@ -732,65 +728,57 @@ checkStepRC x = case x of
 	1 -> return NeedsMore
 	_ -> E.throwIO (SASLException (cToError x))
 
--- | Encode data according to the negotiated SASL mechanism. This might mean
--- the data is integrity or privacy protected.
-encode :: B.ByteString -> Session B.ByteString
-encode input = do
+encodeDecodeHelper :: (F.Storable a, Integral a, Num t) =>
+	   (F.Ptr SessionCtx -> F.Ptr F.CChar -> t -> F.Ptr (F.Ptr F.CChar) -> F.Ptr a -> IO F.CInt)
+	-> B.ByteString
+	-> Session B.ByteString
+encodeDecodeHelper f input = do
 	sctx <- getSessionContext
 	liftIO $
 		B.unsafeUseAsCStringLen input $ \(cstr, cstrLen) ->
 		F.alloca $ \pOutput ->
 		F.alloca $ \pOutputLen -> do
-			rc <- gsasl_encode sctx cstr (fromIntegral cstrLen) pOutput pOutputLen
+			rc <- f sctx cstr (fromIntegral cstrLen) pOutput pOutputLen
 			when (rc /= 0) $ checkCallbackException sctx
 			checkRC rc
 			output <- F.peek pOutput
-			outputLen <- fromIntegral `fmap` F.peek pOutputLen
-			outBytes <- B.packCStringLen (output, outputLen)
+			outputLen <- fromIntegral <$> F.peek pOutputLen
+			outputBytes <- B.packCStringLen (output, outputLen)
 			gsasl_free output
-			return outBytes
+			return outputBytes
+
+-- | Encode data according to the negotiated SASL mechanism. This might mean
+-- the data is integrity or privacy protected.
+encode :: B.ByteString -> Session B.ByteString
+encode = encodeDecodeHelper gsasl_encode
 
 -- | Decode data according to the negotiated SASL mechanism. This might mean
 -- the data is integrity or privacy protected.
 decode :: B.ByteString -> Session B.ByteString
-decode input = do
-	sctx <- getSessionContext
-	liftIO $
-		B.unsafeUseAsCStringLen input $ \(cstr, cstrLen) ->
-		F.alloca $ \pOutput ->
-		F.alloca $ \pOutputLen -> do
-			rc <- gsasl_decode sctx cstr (fromIntegral cstrLen) pOutput pOutputLen
-			when (rc /= 0) $ checkCallbackException sctx
-			checkRC rc
-			output <- F.peek pOutput
-			outputLen <- fromIntegral `fmap` F.peek pOutputLen
-			outputBytes <- B.packCStringLen (output, outputLen)
-			gsasl_free output
-			return outputBytes
+decode = encodeDecodeHelper gsasl_decode
 
 -- }}}
 
 -- Bundled codecs {{{
 
-toBase64 :: B.ByteString -> B.ByteString
-toBase64 input = unsafePerformIO $
+base64Helper :: (F.Storable a, Integral a, Num t) =>
+	   (F.Ptr F.CChar -> t -> F.Ptr (F.Ptr F.CChar) -> F.Ptr a -> IO F.CInt)
+	-> B.ByteString
+	-> B.ByteString
+base64Helper f input = unsafePerformIO $
 	B.unsafeUseAsCStringLen input $ \(pIn, inLen) ->
 	F.alloca $ \pOut ->
 	F.alloca $ \pOutLen -> do
-	gsasl_base64_to pIn (fromIntegral inLen) pOut pOutLen >>= checkRC
+	f pIn (fromIntegral inLen) pOut pOutLen >>= checkRC
 	outLen <- F.peek pOutLen
 	outPtr <- F.peek pOut
 	B.packCStringLen (outPtr, fromIntegral outLen)
 
+toBase64 :: B.ByteString -> B.ByteString
+toBase64 = base64Helper gsasl_base64_to
+
 fromBase64 :: B.ByteString -> B.ByteString
-fromBase64 input = unsafePerformIO $
-	B.unsafeUseAsCStringLen input $ \(pIn, inLen) ->
-	F.alloca $ \pOut ->
-	F.alloca $ \pOutLen -> do
-	gsasl_base64_from pIn (fromIntegral inLen) pOut pOutLen >>= checkRC
-	outLen <- F.peek pOutLen
-	outPtr <- F.peek pOut
-	B.packCStringLen (outPtr, fromIntegral outLen)
+fromBase64 = base64Helper gsasl_base64_from
 
 md5 :: B.ByteString -> B.ByteString
 md5 input = unsafePerformIO $
@@ -804,11 +792,13 @@ md5 input = unsafePerformIO $
 sha1 :: B.ByteString -> B.ByteString
 sha1 input = unsafePerformIO $
 	B.unsafeUseAsCStringLen input $ \(pIn, inLen) ->
-	F.alloca $ \pOut ->
-	F.allocaBytes 20 $ \outBuf -> do
-	F.poke pOut outBuf
+	F.alloca $ \pOut -> do
 	gsasl_sha1 pIn (fromIntegral inLen) pOut >>= checkRC
-	B.packCStringLen (outBuf, 20)
+	outBuf <- F.peek pOut
+	ret <- B.packCStringLen (outBuf, 20)
+	F.free outBuf
+	return ret
+
 
 hmacMD5 :: B.ByteString -- ^ Key
         -> B.ByteString -- ^ Input data
@@ -856,13 +846,6 @@ checkRC x = case x of
 	0 -> return ()
 	_ -> E.throwIO (SASLException (cToError x))
 
-unfoldrM :: Monad m => (b -> m (Maybe (a, b))) -> b -> m [a]
-unfoldrM m b = m b >>= \x -> case x of
-	Just (a, new_b) -> do
-		as <- unfoldrM m new_b
-		return $ a : as
-	Nothing -> return []
-
 -- }}}
 
 -- FFI imports {{{
